/* $Id: civ2.c,v 1.4 2008/11/10 03:39:34 bjj Exp bjj $
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <unistd.h>
#include <string.h>
#include <sys/param.h>
#include <gtk/gtk.h>
#include <libgimp/gimp.h>
#include <libgimp/gimpcompat.h>
#include "civmap.h"

static struct civ2_load_options {
	int	layer_per_civ;
	int	layer_per_improvement;

	char	*image_terrain1;
	char	*image_terrain2;
	char	*image_cities;
	char	*image_units;

	int	show_irrigation;
	int	show_roads;
	int	show_mining;
	int	show_fortresses;
	int	show_units;
	int	show_cities;
} options;

static struct civ2_imagemap {
	gint32		image_ID;
	GimpDrawable	*drawable;
	guchar		*cmap;
	guint		ncolors;
	guchar		cmap_xlate[256];
} terrain1, terrain2, cities, units;

static struct civ2_output_layers {
	GimpDrawable	*drawable;
} output[MAX_TRIBES];

#define gtk_toggle_button_get_state(b)	((b)->active)

/*
 * Look for the Civ2 install dir to provide a useful default location
 * for the imagemaps.  Kind of a pain to fix in the dialog if we're
 * wrong...
 */
static char *
pick_default_image_path()
{
	static char *path[] = {
		"/dos/mps/civ2/",		/* typical install */
		"/msdos/mps/civ2/",		/* typical install */
		"/win/mps/civ2/",		/* typical install */
		"/win95/mps/civ2/",		/* typical install */
		"/cdrom/civ2/",			/* dist cd mounted */
		"/cd/civ2/",			/* dist cd mounted */
		"/media/cdrom0/civ2/",		/* dist cd mounted */
		"/media/cdrom1/civ2/",		/* dist cd mounted */
		"",				/* current directory */
		NULL
	};
	int i;

	for (i = 0; path[i] != NULL; ++i) {
		char name[MAXPATHLEN];

		strcpy(name, path[i]);
		strcat(name, "terrain1.gif");
		if (access(name, R_OK) == 0)
			return path[i];
	}
	/*
	 * What to choose if they can't be found?  Leaving off the path
	 * would be easist for users, but let's hint at what we're looking
	 * for...
	 */
	return path[0];
}

static char *
default_image_path(char *basename)
{
	static char name[MAXPATHLEN];
	static char *path;

	if (!path)
		path = pick_default_image_path();

	strcpy(name, path);
	strcat(name, basename);
	return name;
}

static int dialog_ok = FALSE;

static void
options_cancel_cb(GtkWidget *widget, gpointer data)
{
	dialog_ok = FALSE;
	gtk_main_quit ();
}

static void
options_windelete_cb(GtkWidget *widget, gpointer data)
{
	dialog_ok = FALSE;
}

static void
options_close_cb(GtkWidget *widget, gpointer data)
{
	gtk_main_quit();
}

static void
options_ok_cb(GtkWidget *widget, gpointer data)
{
	dialog_ok = TRUE;
	gtk_main_quit ();
}

static gint
options_dialog()
{
	GtkWidget *dlg;
	GtkWidget *button;
	GtkWidget *frame;
	GtkWidget *vbox;
	GtkWidget *table;
	GtkWidget *label;
	GtkWidget *w_l_c, *w_l_i;
	GtkWidget *w_t1, *w_t2, *w_u, *w_c;
	GtkWidget *w_s_i, *w_s_r, *w_s_m, *w_s_f, *w_s_u, *w_s_c;

	dlg = gtk_dialog_new();
	gtk_window_set_title(GTK_WINDOW(dlg), "Civilization II Saved Game");
	gtk_window_position(GTK_WINDOW(dlg), GTK_WIN_POS_MOUSE);
	gtk_signal_connect(GTK_OBJECT(dlg), "destroy",
				(GtkSignalFunc) options_close_cb,
				NULL);
	gtk_signal_connect(GTK_OBJECT(dlg), "delete_event",
				(GtkSignalFunc) options_windelete_cb,
				NULL);

	button = gtk_button_new_with_label("Ok");
	GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
	gtk_signal_connect(GTK_OBJECT(button), "clicked",
				(GtkSignalFunc) options_ok_cb,
				dlg);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dlg)->action_area), button,
				TRUE, TRUE, 0);
	gtk_widget_grab_default(button);
	gtk_widget_show(button);

	button = gtk_button_new_with_label("Cancel");
	GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
	gtk_signal_connect(GTK_OBJECT(button), "clicked",
				(GtkSignalFunc) options_cancel_cb,
				dlg);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dlg)->action_area), button,
				TRUE, TRUE, 0);
	gtk_widget_show(button);

	frame = gtk_frame_new("Layers");
	gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_ETCHED_IN);
	gtk_container_border_width(GTK_CONTAINER(frame), 10);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dlg)->vbox), frame,
				TRUE, TRUE, 0);
	vbox = gtk_vbox_new(TRUE, 2);
	gtk_container_add(GTK_CONTAINER(frame), vbox);

	w_l_c = gtk_check_button_new_with_label("One set for each Civilization");
	gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(w_l_c), FALSE);
	gtk_box_pack_start(GTK_BOX(vbox), w_l_c, TRUE, TRUE, 0);
	gtk_widget_show(w_l_c);

	w_l_i = gtk_check_button_new_with_label("One for each improvement");
	gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(w_l_i), FALSE);
	gtk_box_pack_start(GTK_BOX(vbox), w_l_i, TRUE, TRUE, 0);
	gtk_widget_show(w_l_i);

	gtk_widget_show(vbox);
	gtk_widget_show(frame);

	frame = gtk_frame_new("Imagemaps");
	gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_ETCHED_IN);
	gtk_container_border_width(GTK_CONTAINER(frame), 10);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dlg)->vbox), frame,
				TRUE, TRUE, 0);
	table = gtk_table_new(4, 2, FALSE);
	gtk_table_set_row_spacings(GTK_TABLE(table), 2);
	gtk_table_set_col_spacings(GTK_TABLE(table), 2);
	gtk_container_add(GTK_CONTAINER(frame), table);

	label = gtk_label_new("TERRAIN1.GIF:");
	gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
	gtk_table_attach(GTK_TABLE(table), label, 0, 1, 0, 1,
			GTK_FILL, GTK_FILL, 0, 0);
	gtk_widget_show(label);
	w_t1 = gtk_entry_new();
	gtk_widget_set_usize(w_t1, 250, 0);
	gtk_entry_set_text(GTK_ENTRY(w_t1),default_image_path("terrain1.gif"));
	gtk_table_attach(GTK_TABLE(table), w_t1, 1, 2, 0, 1,
			GTK_FILL | GTK_EXPAND, 0, 0, 0);
	gtk_widget_show(w_t1);

	label = gtk_label_new("TERRAIN2.GIF:");
	gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
	gtk_table_attach(GTK_TABLE(table), label, 0, 1, 1, 2,
			GTK_FILL, GTK_FILL, 0, 0);
	gtk_widget_show(label);
	w_t2 = gtk_entry_new();
	gtk_widget_set_usize(w_t2, 250, 0);
	gtk_entry_set_text(GTK_ENTRY(w_t2),default_image_path("terrain2.gif"));
	gtk_table_attach(GTK_TABLE(table), w_t2, 1, 2, 1, 2,
			GTK_FILL | GTK_EXPAND, 0, 0, 0);
	gtk_widget_show(w_t2);

	label = gtk_label_new("CITIES.GIF:");
	gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
	gtk_table_attach(GTK_TABLE(table), label, 0, 1, 2, 3,
			GTK_FILL, GTK_FILL, 0, 0);
	gtk_widget_show(label);
	w_c = gtk_entry_new();
	gtk_widget_set_usize(w_c, 250, 0);
	gtk_entry_set_text(GTK_ENTRY(w_c), default_image_path("cities.gif"));
	gtk_table_attach(GTK_TABLE(table), w_c, 1, 2, 2, 3,
			GTK_FILL | GTK_EXPAND, 0, 0, 0);
	gtk_widget_show(w_c);

	label = gtk_label_new("UNITS.GIF:");
	gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
	gtk_table_attach(GTK_TABLE(table), label, 0, 1, 3, 4,
			GTK_FILL, GTK_FILL, 0, 0);
	gtk_widget_show(label);
	w_u = gtk_entry_new();
	gtk_widget_set_usize(w_u, 250, 0);
	gtk_entry_set_text(GTK_ENTRY(w_u), default_image_path("units.gif"));
	gtk_table_attach(GTK_TABLE(table), w_u, 1, 2, 3, 4,
			GTK_FILL | GTK_EXPAND, 0, 0, 0);
	gtk_widget_show(w_u);

	gtk_widget_show(table);
	gtk_widget_show(frame);

	frame = gtk_frame_new("Render");
	gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_ETCHED_IN);
	gtk_container_border_width(GTK_CONTAINER(frame), 10);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dlg)->vbox), frame,
				TRUE, TRUE, 0);
	table = gtk_table_new(2, 3, TRUE);
	gtk_table_set_row_spacings(GTK_TABLE(table), 2);
	gtk_table_set_col_spacings(GTK_TABLE(table), 2);
	gtk_container_add(GTK_CONTAINER(frame), table);

	w_s_i = gtk_check_button_new_with_label("Irrigation");
	gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(w_s_i), TRUE);
	gtk_table_attach(GTK_TABLE(table), w_s_i, 0, 1, 0, 1,
			GTK_FILL, GTK_FILL, 0, 0);
	gtk_widget_show(w_s_i);

	w_s_r = gtk_check_button_new_with_label("Roads");
	gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(w_s_r), TRUE);
	gtk_table_attach(GTK_TABLE(table), w_s_r, 0, 1, 1, 2,
			GTK_FILL, GTK_FILL, 0, 0);
	gtk_widget_show(w_s_r);

	w_s_m = gtk_check_button_new_with_label("Mining");
	gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(w_s_m), TRUE);
	gtk_table_attach(GTK_TABLE(table), w_s_m, 1, 2, 0, 1,
			GTK_FILL, GTK_FILL, 0, 0);
	gtk_widget_show(w_s_m);

	w_s_f = gtk_check_button_new_with_label("Fortresses");
	gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(w_s_f), TRUE);
	gtk_table_attach(GTK_TABLE(table), w_s_f, 1, 2, 1, 2,
			GTK_FILL, GTK_FILL, 0, 0);
	gtk_widget_show(w_s_f);

	w_s_u = gtk_check_button_new_with_label("Units");
	gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(w_s_u), TRUE);
	gtk_table_attach(GTK_TABLE(table), w_s_u, 2, 3, 0, 1,
			GTK_FILL, GTK_FILL, 0, 0);
	gtk_widget_show(w_s_u);

	w_s_c = gtk_check_button_new_with_label("Cities");
	gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(w_s_c), TRUE);
	gtk_table_attach(GTK_TABLE(table), w_s_c, 2, 3, 1, 2,
			GTK_FILL, GTK_FILL, 0, 0);
	gtk_widget_show(w_s_c);

	gtk_widget_show(table);
	gtk_widget_show(frame);
	gtk_widget_show(dlg);

	gtk_main();
	gdk_flush();

	if (dialog_ok) {
		/*
		 * Ok, this isn't stunningly beautiful, but 1) cloning strings
		 * each time the user types in the entry is wrong, and 2) this
		 * is a lot less code than a boatload of callbacks.
		 */
		options.image_terrain1 =
			g_strdup(gtk_entry_get_text(GTK_ENTRY(w_t1)));
		options.image_terrain2 =
			g_strdup(gtk_entry_get_text(GTK_ENTRY(w_t2)));
		options.image_cities =
			g_strdup(gtk_entry_get_text(GTK_ENTRY(w_c)));
		options.image_units =
			g_strdup(gtk_entry_get_text(GTK_ENTRY(w_u)));

		options.layer_per_civ =
			gtk_toggle_button_get_state(GTK_TOGGLE_BUTTON(w_l_c));
		options.layer_per_improvement =
			gtk_toggle_button_get_state(GTK_TOGGLE_BUTTON(w_l_i));

		options.show_irrigation =
			gtk_toggle_button_get_state(GTK_TOGGLE_BUTTON(w_s_i));
		options.show_roads =
			gtk_toggle_button_get_state(GTK_TOGGLE_BUTTON(w_s_r));
		options.show_mining =
			gtk_toggle_button_get_state(GTK_TOGGLE_BUTTON(w_s_m));
		options.show_fortresses =
			gtk_toggle_button_get_state(GTK_TOGGLE_BUTTON(w_s_f));
		options.show_units =
			gtk_toggle_button_get_state(GTK_TOGGLE_BUTTON(w_s_u));
		options.show_cities =
			gtk_toggle_button_get_state(GTK_TOGGLE_BUTTON(w_s_c));
	}

	return dialog_ok;
}


static void
civ2_city_text(gint32 layer, int x, int y, guchar *color, char *text)
{
	gint32 image, floating_sel;
	GimpParam *params;
	int nreturn_vals;
	GimpRGB black;

	gimp_rgb_parse_hex(&black, "#000000", -1);

	image = gimp_drawable_image_id(layer);

	params = gimp_run_procedure("gimp_text_get_extents", &nreturn_vals,
		GIMP_PDB_STRING, text,
		GIMP_PDB_FLOAT, (gfloat) 24.0,	/* height */
		GIMP_PDB_INT32, (gint32) 0,	/* use pixelsize */
		GIMP_PDB_STRING, "*",		/* foundry */
		GIMP_PDB_STRING, "times",		/* family */
		GIMP_PDB_STRING, "medium",		/* weight */
		GIMP_PDB_STRING, "r",		/* slant */
		GIMP_PDB_STRING, "*",		/* set_width */
		GIMP_PDB_STRING, "*",		/* spacing */
		GIMP_PDB_STRING, "*",		/* registry */
		GIMP_PDB_STRING, "*",		/* encoding */
		GIMP_PDB_END);
	if (params[0].data.d_status != GIMP_PDB_SUCCESS)
		return;
	/* center */
	x -= params[1].data.d_int32 / 2;
	y -= params[2].data.d_int32 / 3;
	gimp_destroy_params(params, nreturn_vals);

	gimp_palette_set_foreground(&black);
	params = gimp_run_procedure("gimp_text", &nreturn_vals,
		GIMP_PDB_IMAGE, image,
		GIMP_PDB_DRAWABLE, layer,
		GIMP_PDB_FLOAT, (gfloat)x + 1,
		GIMP_PDB_FLOAT, (gfloat)y + 1,
		GIMP_PDB_STRING, text,
		GIMP_PDB_INT32, (gint32) 1,	/* border */
		GIMP_PDB_INT32, (gint32) 0,	/* antialias */
		GIMP_PDB_FLOAT, (gfloat) 24.0,	/* height */
		GIMP_PDB_INT32, (gint32) 0,	/* use pixelsize */
		GIMP_PDB_STRING, "*",		/* foundry */
		GIMP_PDB_STRING, "times",	/* family */
		GIMP_PDB_STRING, "medium",	/* weight */
		GIMP_PDB_STRING, "r",		/* slant */
		GIMP_PDB_STRING, "*",		/* set_width */
		GIMP_PDB_STRING, "*",		/* spacing */
		GIMP_PDB_STRING, "*",		/* registry */
		GIMP_PDB_STRING, "*",		/* encoding */
		GIMP_PDB_END);
	gimp_destroy_params(params, nreturn_vals);

	if (color) {
		GimpRGB gcolor;

		gimp_rgb_set_uchar(&gcolor, color[0], color[1], color[2]);
		gimp_palette_set_foreground(&gcolor);
	}
	params = gimp_run_procedure("gimp_text", &nreturn_vals,
		GIMP_PDB_IMAGE, image,
		GIMP_PDB_DRAWABLE, layer,
		GIMP_PDB_FLOAT, (gfloat)x,
		GIMP_PDB_FLOAT, (gfloat)y,
		GIMP_PDB_STRING, text,
		GIMP_PDB_INT32, (gint32) 1,	/* border */
		GIMP_PDB_INT32, (gint32) 0,	/* antialias */
		GIMP_PDB_FLOAT, (gfloat) 24.0,	/* height */
		GIMP_PDB_INT32, (gint32) 0,	/* use pixelsize */
		GIMP_PDB_STRING, "*",		/* foundry */
		GIMP_PDB_STRING, "times",	/* family */
		GIMP_PDB_STRING, "medium",	/* weight */
		GIMP_PDB_STRING, "r",		/* slant */
		GIMP_PDB_STRING, "*",		/* set_width */
		GIMP_PDB_STRING, "*",		/* spacing */
		GIMP_PDB_STRING, "*",		/* registry */
		GIMP_PDB_STRING, "*",		/* encoding */
		GIMP_PDB_END);
	floating_sel = params[1].data.d_int32;
	gimp_destroy_params(params, nreturn_vals);
	/* The first is anchored implicitly by the second text */
	gimp_floating_sel_anchor(floating_sel);
}

static gint32
file_gif_load(char *pathname)
{
	GimpParam *params;
	int nparams;

	params = gimp_run_procedure("file_gif_load", &nparams,
		GIMP_PDB_INT32, GIMP_RUN_NONINTERACTIVE,
		GIMP_PDB_STRING, pathname,
		GIMP_PDB_STRING, pathname,
		GIMP_PDB_END);

	if (params[0].data.d_status == GIMP_PDB_SUCCESS)
		return params[1].data.d_image;
	else
		return 0;
}

static void
civ2_tribe_pixel(int tribe, guchar *rgb)
{
	static guchar tribes[MAX_TRIBES][3];
	static int setup = 0;

	if (!setup) {
		int i;

		/* Get the colors from the flags
		 */
		for (i = 0; i < MAX_TRIBES; ++i) {
			GimpTile *tile;
			guchar *data;
			gint x = 1 + i * 15, y = 423;

			tile = gimp_drawable_get_tile2(cities.drawable, 0,x,y);
			gimp_tile_ref(tile);
			x = x % gimp_tile_width();
			y = y % gimp_tile_height();
			data = tile->data + (y * tile->ewidth + x) * tile->bpp;
			data = cities.cmap + *data * 3;
			tribes[i][0] = data[0];
			tribes[i][1] = data[1];
			tribes[i][2] = data[2];
			gimp_tile_unref(tile, FALSE);
		}
		setup = 1;
	}
	rgb[0] = tribes[tribe][0];
	rgb[1] = tribes[tribe][1];
	rgb[2] = tribes[tribe][2];
}

static void
civ2_xparent_idx_to_alpha(gint32 image_ID, GimpDrawable *drawable, gint cmap_size)
{
	guint row, col;
	gint xparent;

	/*
	 * Last three colors are xparent, border and xparent2, per comments
	 * in the original civ2 imagemaps.  Other color indices are also
	 * just used as markers, but they don't appear within the image
	 * rectangle.  Generally they appear in the borders and as text.
	 * Leaving them opaque makes them easy for the rendering engine to
	 * find.
	 */
	xparent = cmap_size - 3;

	for (row = 0; row < drawable->ntile_rows; ++row) for (col = 0; col < drawable->ntile_cols; ++col) {
		GimpTile *tile = gimp_drawable_get_tile(drawable, 0, row, col);
		guchar *data;
		int i;

		gimp_tile_ref(tile);
		data = tile->data;
		i = tile->ewidth * tile->eheight;
		while (i--) {
			if (data[0] >= xparent)
				data[1] = 0;
			data += 2;	/* better be 2! */
		}
		gimp_tile_unref(tile, TRUE);
	}
}

static gint32
civ2_imagemap_load(struct civ2_imagemap *dest, char *pathname)
{
	gint32 image;
	gint32 *layers, layer;
	gint nlayers;
	GimpDrawable *drawable;
	guchar *cmap;

	/*
	 * Load the GIF (yes, it has to be a GIF, since we work from the
	 * special index information)
	 */
	image = file_gif_load(pathname);
	if (!image)
		return 0;

	/*
	 * Find the only layer, promote it to INDEXEDA so we can make the
	 * specially tagged colors transparent.
	 */
	layers = gimp_image_get_layers(image, &nlayers);
	if (!layers || nlayers != 1) {
		gimp_image_delete(image);
		return 0;
	}
	layer = layers[0];
	g_free(layers);
	gimp_layer_add_alpha(layer);
	drawable = gimp_drawable_get(layer);
	dest->cmap = gimp_image_get_cmap(image, &dest->ncolors);
	civ2_xparent_idx_to_alpha(image, drawable, dest->ncolors);

	dest->image_ID = image;
	dest->drawable = drawable;

	return image;
}

static void
civ2_unload_imagemaps()
{
	if (terrain1.image_ID) {
		gimp_drawable_detach(terrain1.drawable);
		gimp_image_delete(terrain1.image_ID);
	}
	if (terrain2.image_ID) {
		gimp_drawable_detach(terrain2.drawable);
		gimp_image_delete(terrain2.image_ID);
	}
	if (cities.image_ID) {
		gimp_drawable_detach(cities.drawable);
		gimp_image_delete(cities.image_ID);
	}
	if (units.image_ID) {
		gimp_drawable_detach(units.drawable);
		gimp_image_delete(units.image_ID);
	}
}

/* Find the union of all of the input cmaps.  This should be less than
 * 256 colors total.  At the same time, record the mapping from imagemap
 * to global cmap for future reference.
 */
static int
civ2_unify_cmap(guchar *cmap, gint *ncolorsp)
{
	struct civ2_imagemap *input,
		*inputs[] = {&terrain1, &terrain2, &cities, &units};
	int i, j, k;
	int ncolors = 0;

	for (i = 0; i < 4; ++i) {
		input = inputs[i];
		for (j = 0; j < input->ncolors; ++j) {
			for (k = 0; k < ncolors; ++k) {
				if (cmap[k*3+0] == input->cmap[j*3+0] &&
				    cmap[k*3+1] == input->cmap[j*3+1] &&
				    cmap[k*3+2] == input->cmap[j*3+2]) {
					break;
				}
			}
			input->cmap_xlate[j] = k;
			if (k == ncolors) {
				if (ncolors > 255)
					return 0;

				cmap[k*3+0] = input->cmap[j*3+0];
				cmap[k*3+1] = input->cmap[j*3+1];
				cmap[k*3+2] = input->cmap[j*3+2];
				++ncolors;
			}
		}
	}
	*ncolorsp = ncolors;

	return 1;
}

static void
civ2_coords(unsigned short x, unsigned short y, guint *xx, guint *yy, int is_tall)
{
	*xx = x * GIFBLOCK_XSIZE + (y & 1)*GIFBLOCK_XSIZE/2;
	*yy = y * GIFBLOCK_YSIZE / 2 + (is_tall ? 0 : GIFBLOCK_YSIZE/2);
}

static void
civ2_find_marker(struct civ2_imagemap *im, guint xin, guint yin, guint *xout, guint *yout, int idx_offs)
{
	static GimpPixelRgn pr_in;
	static guchar data[(GIFBLOCK_XSIZE+1) * (GIFBLOCK_YSIZE*3/2+1) * 2];
	guchar *s;
	guchar target = im->ncolors + idx_offs;
	guint rowstride, i;

	if (pr_in.x != xin || pr_in.y != yin || pr_in.drawable != im->drawable){
		gimp_pixel_rgn_init(&pr_in, im->drawable, xin, yin,
			GIFBLOCK_XSIZE + 1, GIFBLOCK_YSIZE*3/2 + 1,
			FALSE, FALSE);
		gimp_pixel_rgn_get_rect(&pr_in, data, xin, yin, pr_in.w, pr_in.h);
	}
	rowstride = pr_in.w * 2;

	for (i = 0, s = data+2; i < GIFBLOCK_XSIZE; ++i, s += 2)
		if (*s == target)
			break;
	*xout = i;
	for (i = 0, s = data+rowstride; i < GIFBLOCK_YSIZE*3/2; ++i, s += rowstride)
		if (*s == target)
			break;
	*yout = i;
}

static void
civ2_copy_region_raw(GimpDrawable *out, guint xout, guint yout, struct civ2_imagemap *im, guint xin, guint yin, guint width, guint height)
{
	GimpPixelRgn pr_in, pr_out;
	static guchar data[1024 * 2];
	static guchar dest[1024 * 2];
	guchar *s, *d;
	register int i;

	if (width * height > 1024)
		return;

	gimp_pixel_rgn_init(&pr_in, im->drawable, xin, yin, width, height,
		FALSE, FALSE);
	gimp_pixel_rgn_get_rect(&pr_in, data, xin, yin, width, height);

	gimp_pixel_rgn_init(&pr_out, out, xout, yout, width, height,
		TRUE, FALSE);
	gimp_pixel_rgn_get_rect(&pr_out, dest, xout, yout, width, height);

	i = width * height;
	s = data;
	d = dest;
	while (i--) {
		/* we perform kind of a threshold op on the alpha channel
		 * rather than applying it by multiplying...
		 */
		if (s[1]) {
			d[0] = im->cmap_xlate[s[0]];
			d[1] = 255;
		}
		d += 2;
		s += 2;
	}
	gimp_pixel_rgn_set_rect(&pr_out, dest, xout, yout, width, height);
}

static void
civ2_copy_region(GimpDrawable *out, unsigned short x, unsigned short y, struct civ2_imagemap *im, guint xin, guint yin, int is_tall)
{
	guint xout, yout;
	static GimpPixelRgn pr_in, pr_out;
	static guchar data[GIFBLOCK_XSIZE * GIFBLOCK_YSIZE*3/2 * 2];
	static guchar dest[GIFBLOCK_XSIZE * GIFBLOCK_YSIZE*3/2 * 2];
	register guchar *s, *d;
	register int i;

	/*
	 * Assume anything copied from the same place is the same size...
	 */
	if (pr_in.x != xin || pr_in.y != yin || pr_in.drawable != im->drawable){
		gimp_pixel_rgn_init(&pr_in, im->drawable, xin, yin,
			GIFBLOCK_XSIZE,
			is_tall ? GIFBLOCK_YSIZE*3/2 : GIFBLOCK_YSIZE,
			FALSE, FALSE);
		gimp_pixel_rgn_get_rect(&pr_in, data, pr_in.x, pr_in.y,
			pr_in.w, pr_in.h);
	}

	civ2_coords(x, y, &xout, &yout, is_tall);
	if (pr_out.x != xout || pr_out.y != yout || pr_out.drawable != out) {
		gimp_pixel_rgn_init(&pr_out, out, xout, yout, pr_in.w, pr_in.h,
			TRUE, FALSE);
		gimp_pixel_rgn_get_rect(&pr_out, dest, pr_out.x, pr_out.y,
			pr_out.w, pr_out.h);
	}

	i = pr_in.w * pr_in.h;
	s = data;
	d = dest;
	while (i--) {
		/* we perform kind of a threshold op on the alpha channel
		 * rather than applying it by multiplying...
		 */
		if (s[1]) {
			d[0] = im->cmap_xlate[s[0]];
			d[1] = 255;
		}
		d += 2;
		s += 2;
	}
	gimp_pixel_rgn_set_rect(&pr_out, dest, pr_out.x, pr_out.y,
		pr_out.w, pr_out.h);
}

#define ALL_TRIBES -1

static void
civ2_render_city_names(Civmap *map, GimpDrawable *drawable, int tribe)
{
	int i;
	guint32 layer = drawable->drawable_id;

	for (i = 0; i < map->ncities; ++i) {
		guchar rgb[3];
		CivmapCity *city = map->cities + i;
		guint x, y, n;

		if (tribe > 0 && city->tribe != tribe)
			continue;

		civ2_coords(city->x, city->y, &x, &y, 0);
		/* center below city */
		x += GIFBLOCK_XSIZE / 2;
		y += GIFBLOCK_YSIZE;
		civ2_tribe_pixel(city->tribe, rgb);
		civ2_city_text(layer, x, y, rgb, city->name);
	}
}

static void
civ2_render_cities(Civmap *map, GimpDrawable *drawable, int tribe)
{
	int i;
	CivmapCity *city;

	gimp_progress_init("Rendering cities...");
	for (i = 0, city = map->cities; i < map->ncities; ++i, ++city) {
		guchar rgb[3];
		guint xin, yin, n, style;

		if (tribe > 0 && city->tribe != tribe)
			continue;

		gimp_progress_update((double)i / map->ncities);
		/*
		 * choose city style arbitrarily.  should find
		 * the player's chosen style and figure out if
		 * we're post-industrial or modern
		 */
		switch (city->tribe) {
		case 2: case 4:			style = 0; break;
		case 1: case 6:			style = 1; break;
		case 7:				style = 2; break;
		case 3: case 5: 		style = 3; break;
		case 0:				style = 2; break;
		}
		switch (city->pop) {
		case 1: case 2: case 3:		n = 0; break;
		case 4: case 5:			n = 1; break;
		case 6: case 7:			n = 2; break;
		default:			n = 3; break;
		}
		xin = 1 + (GIFBLOCK_XSIZE+1) * n;
		yin = 39 + (GIFBLOCK_YSIZE*3/2+1) * style;
		civ2_copy_region(drawable, city->x, city->y, &cities,
			xin, yin, TRUE);
		/* size */
		/*
		 * occupation flag
		 */
		if (map->data[city->y][city->x].hasUnit) {
			guint xoffs, yoffs, xout, yout;

			civ2_find_marker(&cities, xin-1, yin-1,
				&xoffs, &yoffs, -6);
			yoffs -= 17;	/* XXX mark at base? */
			civ2_coords(city->x, city->y, &xout, &yout, TRUE);
			civ2_copy_region_raw(drawable, xout+xoffs, yout+yoffs,
				&cities, 1 + city->tribe*15, 425, 14, 21);
		}

	}
}

/* the x coordinate of the square left of us is the SAME
 * as our xcoordinate if we're on an odd row, or one over
 * if we're an even row
 */
#define XLEFT(x, y) ((x) - 1 + ((y) & 1))

#define BLK(Mx, My) (map->data[(My)][((Mx) < 0) ? (Mx) + map->xsize : (Mx)])

/* Render roads and railroads
 */
static void
civ2_render_roads(Civmap *map, GimpDrawable *drawable)
{
	unsigned x, y;

	gimp_progress_init("Rendering roads...");
	for (y = 0; y < map->ysize; ++y, gimp_progress_update((double)y / map->ysize)) for (x = 0; x < map->xsize; ++x) {
		CivmapBlock *here = &map->data[y][x];
		int xleft = XLEFT(x, y);
		int i;
		struct {
			int x, y;
		} surround[] = {
			{ xleft + 1, y - 1 },
			{ x + 1, y },
			{ xleft + 1, y + 1 },
			{ x, y + 2},
			{ xleft, y + 1 },
			{ x - 1, y },
			{ xleft, y - 1 },
			{ x, y - 2}
		}, *s;

		/*
		 * roads
		 */
		if (here->hasRoad) {
			int found = 0;
			int railroad = here->hasRailroad;
			guint yin;

			yin = 1 + 11*(GIFBLOCK_YSIZE+1);

			/*
			 * draw a connecting segment toward any adjacent
			 * square that also has a road.  However, if both
			 * this square and the adjacent square have railroads,
			 * *don't* draw the segment...
			 */
			for (i = 1, s = surround; i <= 8; ++i, ++s) {
				if (s->y < 0 || s->y >= map->ysize)
					continue;
				if (!BLK(s->x, s->y).hasRoad ||
				    (railroad && BLK(s->x, s->y).hasRailroad))
					continue;
				found = 1;
				civ2_copy_region(drawable, x, y, &terrain1,
					1 + i*(GIFBLOCK_XSIZE+1), yin, FALSE);
			}
			/*
			 * if there are no adjoining roads there's a special
			 * "crossroad" picture to indicate this square has
			 * a road.  on the other hand, if we have a railroad,
			 * we may just be surrounded by other railroads.
			 * also, if it's a city, it has "implicit roads"
			 */
			if (!found && !railroad && !here->hasCity) {
				civ2_copy_region(drawable, x, y, &terrain1,
					1, yin, FALSE);
			}
		}

		/*
		 * railroads
		 */
		if (here->hasRailroad) {
			int found = 0;
			guint yin;

			yin = 1 + 12*(GIFBLOCK_YSIZE+1);

			/*
			 * draw a connecting segment toward any adjacent
			 * square that also has a road.  However, if both
			 * this square and the adjacent square have railroads,
			 * *don't* draw the segment...
			 */
			for (i = 1, s = surround; i <= 8; ++i, ++s) {
				if (s->y < 0 || s->y >= map->ysize)
					continue;
				if (!BLK(s->x, s->y).hasRailroad)
					continue;
				found = 1;
				civ2_copy_region(drawable, x, y, &terrain1,
					1 + i*(GIFBLOCK_XSIZE+1), yin, FALSE);
			}
			/*
			 * if there are no adjoining roads there's a special
			 * "crossroad" picture to indicate this square has
			 * a railroad.
			 * if it's a city, it has "implicit roads"
			 */
			if (!found && !here->hasCity) {
				civ2_copy_region(drawable, x, y, &terrain1,
					1, yin, FALSE);
			}
		}
	}
}

/* Render mines
 */
static void
civ2_render_mines(Civmap *map, GimpDrawable *drawable)
{
	unsigned x, y;

	for (y = 0; y < map->ysize; ++y) for (x = 0; x < map->xsize; ++x) {
		CivmapBlock *here = &map->data[y][x];

		if (here->hasMining && !here->hasIrrigation)
			civ2_copy_region(drawable, x, y, &terrain1,
				1 + (GIFBLOCK_XSIZE+1) * 7,
				1 + (GIFBLOCK_YSIZE+1) * 5,
				FALSE);
	}
}

/* Render irrigation
 */
static void
civ2_render_irrigation(Civmap *map, GimpDrawable *drawable)
{
	unsigned x, y;

	for (y = 0; y < map->ysize; ++y) for (x = 0; x < map->xsize; ++x) {
		CivmapBlock *here = &map->data[y][x];

		/* irrigation and farmland
		*/
		if (here->hasIrrigation)
			civ2_copy_region(drawable, x, y, &terrain1,
				1 + (GIFBLOCK_XSIZE+1)*7,
				1 + (here->hasMining ?4:3) * (GIFBLOCK_YSIZE+1),
				FALSE);
	}
}

/* Render "natural land"
 */
static void
civ2_render_terrain(Civmap *map, GimpDrawable *drawable)
{
	unsigned x, y;

	gimp_progress_init("Rendering base terrain...");
	for (y = 0; y < map->ysize; ++y, gimp_progress_update((double)y / map->ysize)) for (x = 0; x < map->xsize; ++x) {
		CivmapBlock *here = &map->data[y][x];
		int type = here->type;
		/*
		 * the x coordinate of the square left of us is the SAME
		 * as our xcoordinate if we're on an odd row, or one over
		 * if we're an even row
		 */
		int xleft = x - 1 + (y&1);
		int sel;

		/*
		 * Terrain base
		 */
		civ2_copy_region(drawable, x, y, &terrain1,
			1, 1 + type * (GIFBLOCK_YSIZE+1), FALSE);

		/*
		 * some terrain squares have additional "stuff" on them,
		 * each with several alternate images depending on what
		 * identical stuff is on the squares next to them.
		 *
		 * Rivers work the same way, but go on top of any square.
		 */
		switch(type) {
		case T_Forest:		sel = 1; break;
		case T_Mountains:	sel = 2; break;
		case T_Hills:		sel = 3; break;
		default:		sel = 0;
		}
		for (; sel >= 0; sel = (sel == 0) ? -1 : 0) {
			int tmp;

			if (sel == 0 && !here->hasRiver)
				continue;
			/*
			 * the index of the square we want is nice and
			 * mathematical (almost a bitfield).  a bit
			 * tricky for rivers, though.
			 */
#define EQ(My, Mx) (sel ? (BLK(Mx, My).type == type) : (BLK(Mx, My).hasRiver || BLK(Mx, My).type == T_Ocean))

			tmp = ((y > 0 && EQ(y-1, xleft)) * 8) +
				((y < map->ysize-1 && EQ(y+1, xleft)) * 4) +
				((y < map->ysize-1 && EQ(y+1, xleft+1)) * 2) +
				(y > 0 && EQ(y-1, xleft+1));
			civ2_copy_region(drawable, x, y, &terrain2,
				1 + (tmp & 7) * (GIFBLOCK_XSIZE+1),
				1+((tmp >= 8) + (sel+1)*2) * (GIFBLOCK_YSIZE+1),
				FALSE);
		}

		/*
		 * this coast stuff is tricky!  I almost completely screwed
		 * it up thinking the coast bits were *scaled*!  Whew!
		 */
		if (type == T_Ocean) {
			guint yin = 480 - 3*(COASTBLOCK_YSIZE+1);
			guint xx, yy;
			int tmp;
			char nw, n, ne, e, se, s, sw, w;

			civ2_coords(x, y, &xx, &yy, FALSE);
			nw = y < 1 || BLK(xleft, y-1).type != T_Ocean;
			n = y < 2 || BLK(x, y-2).type != T_Ocean;
			ne = y < 1 || BLK(xleft+1, y-1).type != T_Ocean;
			e = BLK(x+1, y).type != T_Ocean;
			se = y > map->ysize - 2 || BLK(xleft+1, y+1).type != T_Ocean;
			s = y > map->ysize - 3 || BLK(x, y+2).type != T_Ocean;
			sw = y > map->ysize - 2 || BLK(xleft, y+1).type != T_Ocean;
			w = BLK(x-1, y).type != T_Ocean;

			/*
			 * Upper corner...
			 */
			tmp = nw + n*2 + ne*4;
			if (tmp)
				civ2_copy_region_raw(drawable,
					xx + COASTBLOCK_XSIZE/2, yy, &terrain2,
					2*tmp*(COASTBLOCK_XSIZE+1) + 1, yin,
					COASTBLOCK_XSIZE, COASTBLOCK_YSIZE);

			/*
			 * Lower corner...
			 */
			yin += COASTBLOCK_YSIZE + 1;
			tmp = se + s*2 + sw*4;
			if (tmp)
				civ2_copy_region_raw(drawable,
					xx + COASTBLOCK_XSIZE/2,
					yy + COASTBLOCK_YSIZE, &terrain2,
					2*tmp*(COASTBLOCK_XSIZE+1) + 1, yin,
					COASTBLOCK_XSIZE, COASTBLOCK_YSIZE);

			/*
			 * Left corner...
			 */
			yin += COASTBLOCK_YSIZE + 1;
			tmp = nw*4 + w*2 + sw;
			if (tmp)
				civ2_copy_region_raw(drawable,
					xx, yy + COASTBLOCK_YSIZE/2, &terrain2,
					2*tmp*(COASTBLOCK_XSIZE+1) + 1, yin,
					COASTBLOCK_XSIZE, COASTBLOCK_YSIZE);

			/*
			 * Right corner...
			 */
			tmp = se*4 + e*2 + ne;
			if (tmp)
				civ2_copy_region_raw(drawable,
					xx + COASTBLOCK_XSIZE,
					yy + COASTBLOCK_YSIZE/2, &terrain2,
					(2*tmp+1)*(COASTBLOCK_XSIZE+1) + 1, yin,
					COASTBLOCK_XSIZE, COASTBLOCK_YSIZE);

			/*
			 * Okay, now that we've drawn the coast, if we're
			 * by a river, we need to add river mouths...
			 */
			yin = 10 * (GIFBLOCK_YSIZE+1) + 1;
			if (y > 0) {
				if (xleft >=0 && map->data[y-1][xleft].hasRiver)
					civ2_copy_region(drawable, x, y,
						&terrain2,
						3*(GIFBLOCK_XSIZE+1)+1, yin,
						FALSE);
				if (xleft+1 < map->xsize && map->data[y-1][xleft+1].hasRiver)
					civ2_copy_region(drawable, x, y,
						&terrain2, 1, yin, FALSE);
			}
			if (y < map->ysize-1) {
				if (xleft >=0 && map->data[y+1][xleft].hasRiver)
					civ2_copy_region(drawable, x, y,
						&terrain2,
						2*(GIFBLOCK_XSIZE+1)+1, yin,
						FALSE);
				if (xleft+1 < map->xsize && map->data[y+1][xleft+1].hasRiver)
					civ2_copy_region(drawable, x, y,
						&terrain2,
						1*(GIFBLOCK_XSIZE+1)+1, yin,
						FALSE);
			}
		}
	}
}

static void
civ2_render_units(Civmap *map, GimpDrawable *drawable, int tribe)
{
	int i;
	CivmapUnit *unit;

	gimp_progress_init("Rendering units...");
	for (i = 0, unit = map->units; i < map->nunits; ++i, ++unit) {
		guchar rgb[3];
		CivmapBlock *here = &map->data[unit->y][unit->x];
		guint xin, yin, xout, yout, xoffs, yoffs, n;

		if (tribe > 0 && unit->tribe != tribe)
			continue;
		if (here->hasCity && unit->isFortified)
			continue; /* invisible */
		if (!here->hasUnit)
			continue; /* we drew something here */
		here->hasUnit = 0;

		gimp_progress_update((double)i / map->nunits);
		/*
		 * Choose unit pic.  Barbarian leaders have unit type
		 * "diplomat" but a different pic.
		 */
		if (unit->tribe == 0 && unit->type == UNIT_DIPLOMAT)
			n = unit->type + 8;
		else
			n = unit->type;
		xin = 1 + (GIFBLOCK_XSIZE + 1) * (n % 9);
		yin = 1 + (GIFBLOCK_YSIZE*3/2 + 1) * (n / 9);

		/*
		 * draw in the unit's shield.  we know where to put
		 * it from marks in the image map.  this should also
		 * add the hitpoint status and the current orders.
		 * this is drawn first because it goes under the unit
		 */
		civ2_find_marker(&units, xin-1, yin-1, &xoffs, &yoffs, -6);
		civ2_coords(unit->x, unit->y, &xout, &yout, TRUE);
		civ2_copy_region_raw(drawable, xout + xoffs, yout+yoffs,
			&units, 599, 1, 12, 20);

		/*
		 * The unit itself (should check unit->isSleeping)
		 */
		civ2_copy_region(drawable, unit->x, unit->y, &units,
			xin, yin, TRUE);

		/*
		 * draw in the "unit created fortifications"
		 */
		if (unit->isFortified)
			civ2_copy_region(drawable, unit->x, unit->y, &cities,
				143, 423, TRUE);
	}
	/*
	 * Put back the hasUnit flags we unset temporarily...
	 */
	for (i = 0, unit = map->units; i < map->nunits; ++i, ++unit)
		map->data[unit->y][unit->x].hasUnit = 1;
}

static gint32
civ2_render(Civmap *map)
{
	gint32 image = 0;
	gint32 layer = 0;
	GimpDrawable *drawable = NULL;
	guint width, height;
	guchar cmap[256 * 3];
	gint ncolors;
	int i;

	/*
	 * Read the imagemaps.  This should have some error popups as
	 * a side effect...
	 */
	if (!(civ2_imagemap_load(&terrain1, options.image_terrain1) &&
	      civ2_imagemap_load(&terrain2, options.image_terrain2) &&
	      civ2_imagemap_load(&cities, options.image_cities) &&
	      civ2_imagemap_load(&units, options.image_units))) {
		goto fail;
	}
	if (!civ2_unify_cmap(cmap, &ncolors))
		goto fail;

	width = GIFBLOCK_XSIZE * map->xsize + GIFBLOCK_XSIZE/2;
	height = GIFBLOCK_YSIZE*map->ysize/2 + GIFBLOCK_YSIZE;
	image = gimp_image_new(width, height, GIMP_INDEXED);
	gimp_undo_push_group_start(image);
	gimp_image_set_cmap(image, cmap, ncolors);

	layer = gimp_layer_new(image, "Terrain", width, height,
		GIMP_INDEXEDA_IMAGE, 100, GIMP_NORMAL_MODE);
#if 1
	gimp_drawable_fill(layer, GIMP_TRANSPARENT_FILL);
	drawable = gimp_drawable_get(layer);
	civ2_render_terrain(map, drawable);
#else
	gimp_drawable_fill(layer, GIMP_WHITE_FILL);
#endif
	gimp_image_add_layer(image, layer, 0);
	gimp_drawable_detach(drawable);

	if (options.show_roads) {
		layer = gimp_layer_new(image, "Roads", width, height,
			GIMP_INDEXEDA_IMAGE, 100, GIMP_NORMAL_MODE);
		gimp_drawable_fill(layer, GIMP_TRANSPARENT_FILL);
		gimp_image_add_layer(image, layer, 0);
		drawable = gimp_drawable_get(layer);
		civ2_render_irrigation(map, drawable);
		civ2_render_roads(map, drawable);
		civ2_render_mines(map, drawable);
		gimp_drawable_detach(drawable);
	}

	if (options.show_cities) {
		layer = gimp_layer_new(image, "Cities", width, height,
			GIMP_INDEXEDA_IMAGE, 100, GIMP_NORMAL_MODE);
		gimp_drawable_fill(layer, GIMP_TRANSPARENT_FILL);
		gimp_image_add_layer(image, layer, 0);
		drawable = gimp_drawable_get(layer);
		civ2_render_cities(map, drawable, ALL_TRIBES);
		gimp_drawable_detach(drawable);

/* XXX move to top */
		layer = gimp_layer_new(image, "City Names", width, height,
			GIMP_INDEXEDA_IMAGE, 100, GIMP_NORMAL_MODE);
		gimp_drawable_fill(layer, GIMP_TRANSPARENT_FILL);
		gimp_image_add_layer(image, layer, 0);
		drawable = gimp_drawable_get(layer);
		civ2_render_city_names(map, drawable, ALL_TRIBES);
		gimp_drawable_detach(drawable);
	}

	if (options.show_units) {
		layer = gimp_layer_new(image, "Units", width, height,
			GIMP_INDEXEDA_IMAGE, 100, GIMP_NORMAL_MODE);
		gimp_drawable_fill(layer, GIMP_TRANSPARENT_FILL);
		gimp_image_add_layer(image, layer, 0);
		drawable = gimp_drawable_get(layer);
		civ2_render_units(map, drawable, ALL_TRIBES);
		gimp_drawable_detach(drawable);
	}

	if (!image) {
	fail:
		if (image)
			gimp_image_delete(image);
		image = 0;
	}

	civ2_unload_imagemaps();
	if (image)
		gimp_undo_push_group_end(image);
	return image;
}


#include <signal.h>
static void
run(const gchar *name, int nparams, const GimpParam *param, int *nreturn_vals, GimpParam **return_vals)
{
	static GimpParam values[2];
	GimpRunModeType run_mode = param[0].data.d_int32;
	gint32 image_id;

signal(SIGSEGV, SIG_DFL);
	*nreturn_vals = 1;
	*return_vals = values;
	values[0].type = GIMP_PDB_STATUS;
	values[0].data.d_status = GIMP_PDB_CALLING_ERROR;

	if (strcmp(name, "file_civ2_load") == 0) {
		gint32 image;
		int argc = 1;
		static char **argv;
		Civmap *map = NULL;
		FILE *game;

		/*
		 * Read the map or game before we ask them for options!
		 */
		game = fopen(param[1].data.d_string, "r");
		if (game) {
			map = readCivmap(game);
			fclose(game);

			argv = g_new(gchar *, 2);
			argv[0] = "civ2";
			argv[1] = NULL;

			gtk_init(&argc, &argv);
			gtk_rc_parse(gimp_gtkrc());
		}

		if (map && options_dialog() && (image = civ2_render(map))) {
			gimp_image_set_filename(image, param[1].data.d_string);
			*nreturn_vals = 2;
			values[0].data.d_status = GIMP_PDB_SUCCESS;
			values[1].type = GIMP_PDB_IMAGE;
			values[1].data.d_image = image;
		} else {
			values[0].data.d_status = GIMP_PDB_EXECUTION_ERROR;
		}
	} else if (strcmp(name, "file_civ2_save") == 0) {
		/* hey, there isn't one of these */
	}
}

static void
query()
{
	static GimpParamDef args[] = {
		{ GIMP_PDB_INT32, "run_mode", "Interactive, non-interactive" },
		{ GIMP_PDB_STRING, "filename","Name of civ2 .SAV or .MP to load"},
		{ GIMP_PDB_STRING, "raw_filename", "The name entered" },
	};
	static GimpParamDef return_vals[] = {
		{ GIMP_PDB_IMAGE, "image", "Output image" },
	};
	int nargs = sizeof(args) / sizeof(args[0]);
	int nreturn_vals = sizeof(return_vals) / sizeof(return_vals[0]);

	gimp_install_procedure("file_civ2_load",
		"Loads Civilization II SAV and MP files",
		"Loads Civilization II SAV and MP files",
		"Ben Jackson",
		"Ben Jackson",
		"1997",
		"<Load>/CIV2",
		NULL,
		GIMP_PLUGIN,
		nargs, nreturn_vals,
		args, return_vals);
	gimp_register_magic_load_handler("file_civ2_load", "sav", "", "0,string,CIVILIZE");
#if 0
	/* Last one overrides */
	gimp_register_magic_load_handler("file_civ2_load", "mp", "", "");
#endif
}

GimpPlugInInfo PLUG_IN_INFO =
{
	NULL,	/* init */
	NULL,	/* quit */
	query,	/* query */
	run,	/* run */
};

MAIN()
