/* $Id: civmap.c,v 1.1 1998/04/25 10:25:31 bj Exp $
 *
 * By Ben Jackson <ben@ben.com>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "civmap.h"

static
char *typenames[] = {
	"desert",
	"prairie",
	"grasslands",
	"forest",
	"hills",
	"mountains",
	"tundra",
	"arctic",
	"swamp",
	"jungle",
	"ocean"
};

static
char *colors[] = {
	"231 167 123",
	"160 124 63",
	"109 126 68",
	"57 92 51",
	"120 11 65",
	"245 245 245",
	"136 146 115",
	"182 182 176",
	"68 89 67",
	"78 100 75",
	"64 92 168",
};

#define TYPE(Mb, Mx, My) ((Mb)[(Mx)*6 + xsz*(My)*6] & 0xf)

Civmap *
readCivmap(FILE *fp)
{
	int is_saved_game = 0;
	char buf[200000];
	int xsz, ysz;
	int i, j;
	Civmap *map;
	CivmapBlock *block;
	CivmapUnit *unit;
	CivmapCity *city;


	if (fread(buf, 0x62, 1, fp) != 1)
		return NULL;

	map = calloc(sizeof(Civmap), 1);

	if (!strncmp(buf, CIVMAGIC, sizeof(CIVMAGIC)-1)) {
		/*
		 * civ2 saved game.
		 */
		short *t;

		is_saved_game = 1;
		t = (short *)buf;
		map->nunits = t[29];
		map->ncities = t[30];
		map->year = t[15];
		map->turn = t[14];
		/* map->level = buf[13]; */

		fseek(fp, 0x3478, SEEK_SET);
		fread(buf, 0xe, 1, fp);

		xsz = t[0] / 2;
		ysz = t[1];
		/* why is this here? */
		assert(xsz * ysz == t[2]);

		fread(buf, xsz * ysz * 7, 1, fp);
	} else {
		xsz = ((short *)buf)[0] / 2;
		ysz = ((short *)buf)[1];
	}

	map->data = calloc(sizeof(CivmapBlock *), ysz);
	block = calloc(sizeof(CivmapBlock), xsz * ysz);
	for (i = 0; i < ysz; ++i, block += xsz)
		map->data[i] = block;

#if 0
	{
		ushort *xp, *yp;

		xp = buf + 14;
		yp = buf + 14 + 21 * sizeof(ushort);
		for (i = 0; i < 21; ++i, ++xp, ++yp) {
			map->civs[i].x = *xp;
			map->civs[i].y = *yp / 2;
		}
	}
#endif

	map->xsize = xsz;
	map->ysize = ysz;
	fread(map->data[0], xsz * ysz * sizeof(CivmapBlock), 1, fp);

	for (i = xsz * ysz, block = map->data[0]; i; --i, ++block)
		if (block->hasCity)
			block->hasRoad = block->hasRailroad = 1;

	if (is_saved_game) {
		int skip;

		/*
		 * after the main map, some kind of bitmap and then a fixed
		 * blob
		 */
		skip = xsz * (((ysz+3)/4)*4) / 4;
		skip += 0x400;
		fread(buf, skip, 1, fp);

		/*
		 * after that, the units
		 */
		map->units = calloc(sizeof(CivmapUnit), map->nunits);
		fread(map->units, map->nunits, sizeof(CivmapUnit), fp);
		for (unit = map->units, i = map->nunits; i; --i, ++unit) {
			unit->x /= 2;
#if 0
			fprintf(stderr, "(%d, %d) %s", unit->x, unit->y, unitnames[unit->type]);
			for (j = 0; j < 11; ++j)
				fprintf(stderr, " %x", unit->dunno[j]);
			fputc('\n', stderr);
#endif
		}

		/*
		 * after that, the cities
		 */
		map->cities = calloc(sizeof(CivmapCity), map->ncities);
		fread(map->cities, map->ncities, sizeof(CivmapCity), fp);
		for (city = map->cities, i = map->ncities; i; --i, ++city) {
			city->x /= 2;
		}
	}
	return map;
}

int
findWater(Civmap *map)
{
	int nocean[256], max, i, j, half;

	for (i = 0; i < map->xsize; ++i) {
		max = 0;
		for (j = 0; j < map->ysize; ++j)
			if (map->data[j][i].type == T_Ocean)
				++max;
		nocean[i] = max;
	}
	for (i = j = half = max = 0; i < map->xsize; ++i) {
		if (nocean[i] > max) {
			max = nocean[i];
			j = i;
			half = 0;
		} else if (nocean[i] == max) {
			++half;
		}
	}
	return j + half/2;
}

#if 0
void
rotateMap(Civmap *map, int offs)
{
	int i, j;
	CivmapBlock *block;
	CivmapUnit *unit;
	CivmapCity *city;

	for (j = 0; j < map->ysz; ++j) {
		block = map->data[j];
		if (offs < 0) {
			memcpy
		}
	}
}
#endif

char *
typeName(int type)
{
	return typenames[type];
}
